# Copyright (c) 2021 Alphons Gwatimba
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT
__version__ = "2022.4.6"

from rainbow.optical_flow.optical_flow import OPTICAL_FLOW_FILENAME

RAW_IMGS_DIR_NAME = 'Raw_Images'
FLOW_IMGS_DIR_NAME = 'Flow_Images'
COMB_IMGS_DIR_NAME = 'Combined_Images'
METADATA_FILENAME = 'metadata'
MAG_STATS_FILENME = 'magnitude_stats'
DIRN_STATS_FILENME = 'direction_stats'
MAG_SCATTER_FILENAME = 'magnitude_scatter_plot'
DIRN_SCATTER_FILENAME = 'direction_scatter_plot'
HEATMAPS_DIR_NAME = 'Heatmaps'
QUIVER_PLOTS_DIR_NAME = 'Quiver_Plots'
POLAR_PLOTS_DIR_NAME = 'Polar_Plots'



from .gma import GMA


class ModelFactory:
    """Factory for initializing optical flow models.
    """
    def get_model(self, model, mdl_cfg, reuse_mdl):
        """Initializes a supported optical flow model.

        Args:
            model (string): The name of the optical flow model.
            mdl_cfg (dict): The configuration of the optical flow model.
            reuse_mdl (bool): If True, multiple calls of this function
                with the same model name will return the same optical model
                instance.

        Raises:
            ValueError: The chosen optical flow model is not supported.

        Returns:
            BaseModel: Supported optical flow model.
        """
        if model == 'gma':
            return GMA.get_instance(mdl_cfg, reuse_mdl)
        else:
            msg = f'Chosen optical flow model ({model}) not supported.'
            raise ValueError(msg)



import os

import cv2

import numpy as np

import rainbow
from rainbow.optical_flow.model_factory import ModelFactory
from rainbow.util import (cleanup_dir, save_img_ser, save_img_ser_metadata,
                          save_optical_flow)

OPTICAL_FLOW_FILENAME = 'optical_flow.npy'


def compute_optical_flow(imgs, output_dir, model_name, model_config,
                         reuse_model=True, save_raw_imgs=True,
                         overwrite_flow=False):
    """Computes the optical flow.

    Computes the optical flow in an image sequence using the chosen optical
    flow model and saves the flow as a file in output_dir.

    Args:
        imgs (list): A list of images.
        output_dir (string): The path to the output directory.
        model_name (string): The name of the optical flow model to use.
        model_config (dict): The configuration of the optical flow model.
        reuse_model (bool, optional): If True, multiple calls of this function
            with the same model name will use the same optical model instance.
            Defaults to True.
        save_raw_imgs (bool, optional): If True, saves the image sequence in
            output_dir. Defaults to True.
        overwrite_flow (bool, optional): If True, will re-compute optical flow
            even if the flow file already exists in output_dir.
            Defaults to False.
    """
    mdl_fcty = ModelFactory()
    model = mdl_fcty.get_model(model_name, model_config, reuse_model)
    preds = model.predict(imgs)

    if not (os.path.isfile(os.path.join(output_dir,
            OPTICAL_FLOW_FILENAME)) and not overwrite_flow):
        if not cleanup_dir(output_dir):
            return
        os.mkdir(output_dir)
        save_optical_flow(preds, output_dir)

        if save_raw_imgs:  # TODO remove
            raw_imgs_dir = os.path.join(output_dir, rainbow.RAW_IMGS_DIR_NAME)
            os.mkdir(raw_imgs_dir)
            save_img_ser(imgs, raw_imgs_dir)
            save_img_ser_metadata(imgs, output_dir)


def flow_to_img(flow, normalize=True, info=None, flow_mag_max=None):
    """Visualizes optical flow (https://github.com/philferriere/tfoptflow).

    Convert flow to viewable image, using color hue to encode flow vector
    orientation, and color saturation to encode vector length. This is similar
    to the OpenCV tutorial on dense optical flow, except that they map vector
    length to the value plane of the HSV color model, instead of the saturation
    plane, as we do here.
    [Ref] OpenCV 3.0.0-dev documentation » OpenCV-Python Tutorials » Video
        Analysis »
        https://docs.opencv.org/3.0-beta/doc/py_tutorials/py_video/
            py_lucas_kanade/py_lucas_kanade.html

    Args:
        flow (numpy array): The optical flow.
        normalize (bool, optional): If True, normalizes flow to 0..255.
            Defaults to True.
        info (string, optional): Text to superimpose on image (typically, the
            epe for the predicted flow). Defaults to None.
        flow_mag_max (int, optional): Max flow to map to 255. Defaults to None.

    Returns:
        numpy array: Viewable representation of the dense optical flow in RGB
            format.
    """
    hsv = np.zeros((flow.shape[0], flow.shape[1], 3), dtype=np.uint8)
    flow_magnitude, flow_angle = (cv2.cartToPolar(flow[..., 0].astype(
                                  float), flow[..., 1].astype(float)))

    # A couple times, we've gotten NaNs out of the above...
    nans = np.isnan(flow_magnitude)
    if np.any(nans):
        nans = np.where(nans)
        flow_magnitude[nans] = 0.

    # Normalize.
    hsv[..., 0] = flow_angle * 180 / np.pi / 2
    if normalize is True:
        if flow_mag_max is None:
            hsv[..., 1] = cv2.normalize(flow_magnitude, None, 0, 255,
                                        cv2.NORM_MINMAX)
        else:
            hsv[..., 1] = flow_magnitude * 255 / flow_mag_max
    else:
        hsv[..., 1] = flow_magnitude
    hsv[..., 2] = 255
    img = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB)

    # Add text to the image, if requested.
    if info is not None:
        font = cv2.FONT_HERSHEY_SIMPLEX
        cv2.putText(img, info, (20, 20), font, 0.8, (0, 0, 0), 2, cv2.LINE_AA)

    return img



import os
import sys
import urllib
import warnings
from argparse import Namespace

import imutils

from rainbow.optical_flow.base_model import BaseModel

MIN_DIMS = (284, 121)
CHECKPOINTS_BASE_URL = 'https://github.com/AlphonsG/GMA/raw/main/checkpoints/'


class GMA(BaseModel):
    """Optical flow model from "Learning to Estimate Hidden Motions with Global
    Motion Aggregation" by Jiang et. al (https://github.com/zacjiang/GMA).
    """
    __instance__ = None

    def __init__(self, model_cfg, reuse_model):
        """See base class."""
        gma_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                               'third_party', 'gma', 'core')
        sys.path.insert(1, gma_dir)
        from network import RAFTGMA
        from utils.utils import InputPadder
        import torch
        self.InputPadder = InputPadder
        self.torch = torch
        args = Namespace(**model_cfg)
        model = self.torch.nn.DataParallel(RAFTGMA(args))

        if os.path.isfile(model_cfg['model']):
            model.load_state_dict(self.torch.load(model_cfg['model'],
                                  map_location=self.torch.device(
                                      model_cfg['device'])))
        else:
            checkpoint_url = CHECKPOINTS_BASE_URL + model_cfg['model']
            try:
                model.load_state_dict(self.torch.hub.load_state_dict_from_url(
                    checkpoint_url, map_location=self.torch.device(
                        model_cfg['device'])))
            except urllib.error.URLError as e:
                msg = ('\nERROR: The optical flow model checkpoint specified '
                       f'in the YAML config file ({model_cfg["model"]}) is '
                       'not a local file, and the attempt to download it '
                       f'from {CHECKPOINTS_BASE_URL} failed ({str(e)}).')
                print(msg)
                exit(1)

        model = model.module
        model.to(model_cfg['device'])
        model.eval()
        self.model = model
        self.model_cfg = model_cfg
        GMA.__instance__ = self

    @staticmethod
    def get_instance(model_cfg, reuse_model):
        """See base class."""
        if GMA.__instance__ is not None:
            if not reuse_model:
                raise ValueError('Cannot create multiple GMA instances.')
        else:
            GMA(model_cfg, reuse_model)

        return GMA.__instance__

    def predict(self, imgs):
        """See base class."""
        imgs = imgs.copy()
        # Resize images to minimum dimensions, if necessary.
        for i, img in enumerate(imgs):
            for j, kv in zip([0, 1], [{'height': MIN_DIMS[0]},
                                      {'width': MIN_DIMS[1]}]):
                if img.shape[j] < MIN_DIMS[j]:
                    img = imutils.resize(img, **kv)
            imgs[i] = img

        imgs = [self.torch.from_numpy(img).permute(2, 0, 1).float()[None].to(
                self.model_cfg['device']) for img in imgs]
        imgs = self.get_img_pairs(imgs)

        # Pad images, if necessary.
        padder = self.InputPadder(imgs[0][0].shape)
        imgs = [(padder.pad(img1, img2)) for img1, img2 in imgs]

        preds = []
        with self.torch.no_grad():
            for img1, img2 in imgs:
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', category=UserWarning)
                    _, flow = self.model(img1, img2, iters=12, test_mode=True)
                preds.append(flow[0].permute(1, 2, 0).cpu().numpy())

        return preds



import os
import sys

from gooey import Gooey, GooeyParser

from rainbow.file_processing import process_files

import yaml


if len(sys.argv) >= 2 and '--ignore-gooey' not in sys.argv:
    sys.argv.append('--ignore-gooey')


@Gooey(hide_progress_msg=True)
def process_args():
    parser = GooeyParser(description='Automated air liquid interface cell '
                         'culture analysis using deep optical flow.')
    parser.add_argument('root_dir', type=str, help='Path to root directory '
                        'to begin searching for image sequences to process '
                        'in', widget='DirChooser')
    parser.add_argument('config', type=str, help='Path to YAML configuration '
                        'file', widget='FileChooser')
    parser.add_argument('--num-workers', type=int, help='Maximum '
                        'number of workers to use for parallel analysis '
                        '(default number equals CPU core count)',
                        widget='IntegerField', gooey_options={'min': 1})
    parser.add_argument('--subdirs', help='Process files in root directory '
                        'subfolders instead', action='store_true')
    parser.add_argument('--overwrite-flow', action='store_true',
                        help='Recompute optical flow even if preexisting '
                             'optical flow file is found for an image series')
    args = parser.parse_args()

    return args


def main():
    # check if pytorch is installed
    try:
        import torch
    except ModuleNotFoundError:
        msg = ('Rainbow cannot run because PyTorch is currently '
               'not installed. Please install Pytorch by following the '
               'instructions here: https://pytorch.org/get-started/locally/.')
        print(msg)
        exit(1)

    args = process_args()
    assert os.path.isdir(args.root_dir), ('Invalid root directory path '
                                          'provided.')
    assert os.path.isfile(args.config), 'Invalid config file path provided.'
    assert args.num_workers is None or args.num_workers > 0, (
        'Invalid number of workers provided.')

    with open(args.config) as f:
        config = yaml.safe_load(f)

    process_files(args.root_dir, config, args.num_workers, args.subdirs,
                  args.overwrite_flow)

    return 0


if __name__ == "__main__":
    __spec__ = None  # pdb multiprocessing support
    main()



import json
import os
import shutil
import tempfile
import warnings
from datetime import datetime
from pathlib import Path

import cv2

from moviepy.editor import ImageSequenceClip

from natsort import natsorted

from nd2reader import ND2Reader

import numpy as np

from pims import Frame

import rainbow

VID_FILE_EXT = '.mp4'


def load_nd2_imgs(nd2, axs_config, mpp=None):  # mpp None?
    """Loads image sequences from a .nd2 file.

    Args:
        nd2 (string): The path to a .nd2 file.
        axs_config (dict): A dictionary with the key 'iter_axs', containing a
            list of the ordered nd2 axes to iterate over, and the key
            'bdl_axs', containing a list of the axes to bundle when iterating.
        mpp (float, optional): The micrometres per pixel value of the
            image sequences, if missing from the .nd2 file. Defaults to None.

    Returns:
        list: A list of image sequences loaded from the .nd2 file.
    """
    frms = ND2Reader(nd2)
    avl_axs = list(frms.sizes.keys())
    iter_axs = [ax for ax in axs_config['iter_axs'] if ax in avl_axs]
    bdl_axs = [ax for ax in axs_config['bdl_axs'] if ax in avl_axs]
    if len(iter_axs) == 0 or len(bdl_axs) == 0:
        return []
    frms.iter_axes, frms.bundle_axes = ''.join(iter_axs), ''.join(bdl_axs)
    min_px, max_px, ser_len = (0, np.array(list(frms)).max(),
                               frms.sizes[iter_axs[-1]])
    imgs, curr_img_ser = [], []
    for i, frm in enumerate(frms):
        img = (frm.astype(float) - min_px) * 255.0 / (max_px - min_px)
        img = np.asarray(img, dtype=np.uint8)
        img = img[:, :, np.newaxis]
        img = np.repeat(img, 3, axis=2)
        frame = Frame(img)
        frame.metadata = frm.metadata
        frame.metadata['type'] = '.nd2'
        frame.metadata['path'] = nd2
        frame.metadata['img_name'] = 'Image_{}.png'.format(
            frm.metadata['coords'][axs_config['iter_axs'][-1]])
        try:
            frame.metadata['mpp'] = frame.metadata['pixel_microns']
        except KeyError:
            frame.metadata['mpp'] = mpp

        curr_img_ser.append(frame)
        if (i + 1) % ser_len == 0:
            curr_img_ser[0].metadata['img_ser_md'] = frms.metadata
            curr_img_ser[0].metadata['img_ser_md']['type'] = '.nd2'
            curr_img_ser[0].metadata['img_ser_md']['dir'] = os.path.dirname(
                nd2)
            imgs.append(curr_img_ser)
            curr_img_ser = []

    frms.close()

    return imgs


def load_std_imgs(input_dir, mpp=None):
    """Loads an image sequence from input_dir.

    Assumes order of image sequence corresponds to naturally sorted image
    filenames in input_dir.

    Args:
        input_dir (string): The path to the input directory.
        mpp (float, optional): The micrometres per pixel value of the
            image sequence. Defaults to None.

    Returns:
        list: A list of images.
    """
    try:
        _, _, files = next(os.walk(input_dir))
    except StopIteration:
        return []

    files = natsorted([os.path.join(input_dir, f) for f in files])
    imgs = []
    for f in files:
        img = cv2.imread(f)
        if img is not None:
            frame = Frame(img)
            frame.metadata['type'], frame.metadata['path'], frame.metadata[
                'img_name'], frame.metadata['mpp'] = (Path(f).suffix, f,
                                                      Path(f).name, mpp)  # mp?

            if len(imgs) == 0:
                frame.metadata['img_ser_md'] = {'type': Path(f).suffix,
                                                'dir': input_dir,
                                                'calibration_um': mpp
                                                }
            imgs.append(frame)

    return imgs


def save_img_ser(imgs, output_dir, use_metadata_name=True):
    """Saves an image sequence to output_dir.

    Saves images in .png format. Automatically names image files or uses
    stored image names in image sequence metadata.

    Args:
        imgs (list): A list of images.
        output_dir (string): The path to the output directory.
        use_metadata_name (bool, optional): If True, names images using image
            names in sequence metadata. Defaults to True.
    """
    for i, img in enumerate(imgs):
        path = (os.path.join(output_dir, img.metadata['img_name']) if
                use_metadata_name else os.path.join(output_dir,
                'Image_{}.png'.format(i)))
        cv2.imwrite(path, img)


def cleanup_dir(output_dir):
    """Cleans up output_dir.

    If output_dir exists, attempts to remove it and its contents.

    Args:
        output_dir (string): The path to a directory.

    Returns:
        bool: If True, output_dir is guaranteed not to exist.
    """
    if os.path.isdir(output_dir):
        try:
            shutil.rmtree(output_dir)
        except OSError as e:
            msg = (f'Could not remove existing directory ({output_dir}), '
                   f'reason: {str(e)}.')
            warnings.warn(msg, UserWarning)
            return False

    return True


def comb_imgs(img1, img2):
    """Horizontally concatenates two images.

    img1 becomes the leftmost image, while img2 becomes the rightmost image.

    Args:
        img1 (numpy.array): An image.
        img2 (numpy.array): An image.

    Returns:
        numpy.array: Image result of horizontally concatenating the two images.
    """
    return np.concatenate((img1, img2), axis=1)


def save_video(input_dir, output_path, fps=5):
    """Saves a video.

    Saves a video generated from an image sequence, located in input_dir,
    to output_path.

    Args:
        input_dir (string): The path to the input directory.
        output_path (string): The path to write the generated video to
            (includes video filename without extension).
        fps (int, optional): The framerate of the generated video. Defaults
            to 5.

    Returns:
        bool: If True, video was successfully saved.
    """
    imgs = load_std_imgs(input_dir)
    if len(imgs) == 0:
        return False

    temp_report = os.path.join(tempfile.gettempdir(), 'temp_report.log')
    os.environ['FFREPORT'] = f'file=\"{temp_report}\"'
    video = ImageSequenceClip([cv2.cvtColor(img, cv2.COLOR_BGR2RGB) for img in
                               imgs], fps=fps)
    video.write_videofile(os.path.join(output_path + VID_FILE_EXT),
                          logger=None, write_logfile=False)
    video.close()

    return True


def apply_metadata(img1, img2):
    """Sets img1's metadata to that of img2.

    Args:
        img1 (PIMS Frame): An image with a 'metadata' attribute.
        img2 (numpy.array): An image.

    Returns:
        PIMS Frame: An image.
    """
    img1 = Frame(img1)
    img1.metadata = img2.metadata

    return img1


def save_optical_flow(preds, output_dir):
    """Saves optical flow predictions to a file in output_dir.

    Args:
        preds (list): A list of optical flow predictions as numpy arrays.
        output_dir (string): The path to the output directory.
    """
    np.save(os.path.join(output_dir, rainbow.OPTICAL_FLOW_FILENAME),
            np.array(preds), allow_pickle=False)


def load_optical_flow(flow_path):
    """Loads optical flow predictions from a file.

    Args:
        flow_path (string): The path to a .npy file containing saved optical
            flow predictions.

    Returns:
        list: A list of optical flow predictions as numpy arrays.
    """
    raw_preds = np.load(flow_path)
    preds = []
    for i in range(0, raw_preds.shape[0]):
        preds.append(raw_preds[i, ...])

    return preds


def save_img_ser_metadata(imgs, output_dir):
    """Saves image sequence metadata to a .json file in output_dir.

    Args:
        imgs (list): A list of images as PIMS Frame arrays.
        output_dir (string): The path to the output directory.
    """
    metadata = imgs[0].metadata['img_ser_md']
    now = datetime.now()
    dt_str = now.strftime('%d %B %Y, %H:%M:%S')
    metadata['analysis_timestamp'] = dt_str
    with open(os.path.join(output_dir, f'{rainbow.METADATA_FILENAME}.json'),
              'w') as f:
        json.dump(metadata, f, indent=4, sort_keys=True, default=str)


def video_reshape(vid_path, set_wdh=None):
    """Resizes a video.

    Calculates the video dimensions of a video, while maintaining the aspect
    ratio, if the width were resized.


    Args:
        vid_path (string): The path to a video file.
        set_wdh (int, optional): The new width of the video. Defaults to None.

    Returns:
        int, int: The new width and height of the video.
    """
    cap = cv2.VideoCapture(vid_path)
    hgt, wdh, _ = cap.read()[1].shape
    dsp_wdh = set_wdh if set_wdh is not None else wdh
    dsp_hgt = dsp_wdh * (hgt / wdh) if wdh is not None else hgt
    cap.release()

    return round(dsp_wdh), round(dsp_hgt)



import csv
import math
import os
import warnings
from itertools import chain
from pathlib import Path

import cv2

import matplotlib.pyplot as plt

import nbconvert
from nbconvert.preprocessors import CellExecutionError, ExecutePreprocessor

import nbformat

import numpy as np

from physt import polar

import scipy.stats

SENTINEL = 'STOP'
NOTEBOOK_DIR = 'misc/notebooks'


def analyze_data(queue, config):
    """Performs data analysis.

    Receives directories from worker processes via queue and generates an
    analysis report of data stored in those directories.

    Args:
        queue (multiprocessing.Queue): The queue that will be used to receive
            directory paths from workers.
        config (string): The path to a .yaml configuration file.

    Returns:
        int: 0 when data analysis has successfully finished.
    """
    while True:
        output_dir = queue.get()
        if output_dir == SENTINEL:
            # put it back so that other consumers see it
            queue.put(SENTINEL)
            return 0

        gen_report(output_dir, config['report_path'])


def gen_report(output_dir, report_path, html=True):
    """Generate a report.

    Execute a jupyter notebook in output_dir.

    Args:
        output_dir (string): The path to the output directory.
        report_path (string): The path to the jupyter notebook.
        html (bool, optional): If True, will also save executed jupyter
            notebook as a html file in output_dir. Defaults to True.
    """
    if not os.path.isfile(report_path):
        report_path = os.path.join(os.path.abspath(os.path.dirname(
            __file__)), '..', NOTEBOOK_DIR, report_path)
        if not os.path.isfile(report_path):
            return

    with open(report_path) as f:
        nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600, kernel_name='python3')
        ep.allow_errors = True
        gend_report_path = os.path.join(output_dir,
                                        Path(report_path).name)
        try:
            ep.preprocess(nb, {'metadata': {'path': output_dir}})
        except CellExecutionError:
            msg = (f'Could not generate report, see \'{gend_report_path}\' '
                   'for error.')
            warnings.warn(msg, UserWarning)
        finally:
            with open(gend_report_path, 'w', encoding='utf-8') as f:
                nbformat.write(nb, f)
            if html:
                save_html(output_dir, gend_report_path)


def save_html(output_dir, gend_report_path):
    """Saves jupyter notebook as a html file in output_dir.

    Args:
        output_dir (string): The path to the output directory.
        gend_report_path (string): The path to the jupyter notebook.
    """
    with open(gend_report_path) as f:
        nb = nbformat.read(f, as_version=4)
        exporter = nbconvert.HTMLExporter()
        exporter.exclude_input = True
        body, resources = exporter.from_notebook_node(nb)
        file_writer = nbconvert.writers.FilesWriter()
        file_writer.write(output=body, resources=resources,
                          notebook_name=os.path.join(output_dir, Path(
                                                     gend_report_path).stem))


def mag_scatter(mags):
    """Creates a magnitude scatter plot.

    Args:
        mags (list): A list of floats representing vector magnitudes.
    """
    times = list(range(1, len(mags) + 1))
    plt.figure()
    plt.scatter(times, mags)
    plt.title('Flow Magnitude')
    plt.xlabel('Frame Number (Frame)')
    plt.ylabel('Average Distance (μm)')
    plt.ylim(0, max(mags) * 1.5)


def dirn_scatter(dirns):
    """Creates a direction scatter plot.

    Args:
        dirns (list): A list of floats representing vector directions.
    """
    times = list(range(1, len(dirns) + 1))
    plt.figure()
    plt.scatter(times, dirns)
    plt.title('Flow Direction')
    plt.xlabel('Frame Number (Frame)')
    plt.ylabel('Average Angle (Deg)')
    plt.ylim(0, max(dirns) * 1.5)


def save_heatmaps(preds, output_dir, step=0, dpi=1000):
    """Saves heatmaps.

    Writes images visualizing the magnitude of optical flow of an image
    sequence to output_dir.

    Args:
        preds (list): The optical flow across an image sequence.
        output_dir (string): The path to the output directory.
        step (int, optional): The amount of infomation discarded when
            visualizing heatmaps. Defaults to 0.
        dpi (int, optional): The dots per inch of the saved heatmap images.
            Defaults to 1000.
    """
    for i, pred in enumerate(preds):
        plt.xticks([])
        plt.yticks([])
        plt.Axes(plt.figure(frameon=False), [0, 0, 1, 1])
        plt.axis('off')
        pred = np.copy(pred)
        pred[:, :, 1] = pred[:, :, 1] * -1
        u, v = ((pred[::step, ::step, 0].astype(np.float32),
                pred[::step, ::step, 1].astype(float)) if step != 0 else
                (pred[:, :, 0].astype(float), pred[:, :, 1].astype(float)))
        mag = cv2.cartToPolar(u, v, angleInDegrees=1)
        plt.pcolormesh(mag[0], cmap='hot')
        path = os.path.join(output_dir, 'Image_{}.png'.format(i))
        plt.savefig(path, dpi=dpi, bbox_inches=0)
        plt.close()


def save_quiver_plots(preds, output_dir, step=70, dpi=1000):
    """Saves quiver plots.

    Writes images visualizing the direction of optical flow of an image
    sequence to output_dir.

    Args:
        preds (list): The optical flow across an image sequence.
        output_dir (string): The path to the output directory.
        step (int, optional): The pixel seperation of the vectors visualized in
            the quiver plots.
        dpi (int, optional): The dots per inch of the saved quiver plot images.
            Defaults to 1000.
    """
    for i, pred in enumerate(preds):
        plt.xticks([])
        plt.yticks([])
        plt.Axes(plt.figure(frameon=False), [0, 0, 1, 1])
        plt.axis('off')
        pred = np.copy(pred)
        pred[:, :, 1] = pred[:, :, 1] * -1
        plt.quiver(np.arange(0, pred.shape[1], step), np.arange(pred.shape[0],
                   -1, -step), pred[::step, ::step, 0],
                   pred[::step, ::step, 1])
        path = os.path.join(output_dir, 'Image_{}.png'.format(i))
        plt.savefig(path, dpi=dpi, bbox_inches=0)
        plt.close(plt.gcf())


def gen_base_metrics(data):
    """Generates metrics for the optical flow data of an image sequence.

    Args:
        data (Defaultdict(list)): Contains an image sequence under the key
            'raw_imgs' and the corresponding optical flow under the key
            'preds'.
    """
    for i, pred in enumerate(data['preds']):
        mag, dirn = cv2.cartToPolar(pred[..., 0].astype(float),
                                    pred[..., 1].astype(float),
                                    angleInDegrees=1)
        data['mags'].append(mag)
        data['dirns'].append(dirn)
        mag[np.isnan(mag)] = 0  # remove NaNs
        mag *= data['raw_imgs'][0].metadata['mpp']  # convert to um, calib?
        for key, metric in zip(['mag_stats', 'dirn_stats'], [mag, dirn]):
            circ = True if key == 'dirn_stats' else False
            data[key].append(gen_stats(metric, i, circ))


def gen_stats(metric, i, circ=False):
    """Calculates statistics for a given metric.

    The statistics (e.g. mean, variance, etc.) are for a given metric in a
    frame of an image sequence.

    Args:
        metric (list): A list of values for the given metric.
        i (int): The corresponding frame number for the values of the given
            metric.
        circ (bool, optional): If true, treats metric as circular data.
            Defaults to False.
    Returns:
        dict: The statistics for the given metric.
    """
    if circ:
        metric = list(chain.from_iterable(metric.tolist()))
        mini, maxi = min(metric), max(metric)
        metric = np.deg2rad(metric)
        mean, std, var = scipy.stats.circmean(metric), scipy.stats.circstd(
            metric), scipy.stats.circvar(metric)
        mean, std, var = np.rad2deg(mean), np.rad2deg(std), np.rad2deg(var)
        stats = {'frame': i, 'min': mini, 'max': maxi, 'mean': mean,
                 'std': std, 'var': var}
    else:
        stats = scipy.stats.describe(metric, axis=None)
        stats = {'frame': i, 'min': stats[1][0],
                 'max': stats[1][1], 'mean': stats[2],
                 'std': math.sqrt(stats[3]), 'var': stats[3]}

    return stats


def save_stats(stats, csv_path, unit):
    """Saves statistics to a csv file.

    The statistics are for a given metric in one or more frames of an image
    sequence.

    Args:
        stats (list): A list of statistics.
        csv_path (string): The path of the csv file to create.
        unit (string): The unit of measurement for the given metric.
    """
    try:
        with open(csv_path, 'w', newline='') as csv_file:
            fields = ['frame', 'min', 'max', 'mean', 'std', 'var']  # outside?
            fields = ['{} ({})'.format(fd, unit) if fd != 'frame' else fd for
                      fd in fields]
            writer = csv.writer(csv_file)
            writer.writerow(fields)
            for s in stats:
                writer.writerow(list(s.values()))
    except (OSError, csv.Error, ValueError) as e:
        msg = 'Could not create stats file. Message: {}.'.format(str(e))
        warnings.warn(msg, UserWarning)


def save_polar_plots(preds, output_dir, dpi=1000, mpp=None):
    """Saves polar plots.

    Writes polar plots visualizing the optical flow circular data of an image
    sequence to output_dir.

    Args:
        preds (list): The optical flow across an image sequence.
        output_dir (string): The path to the output directory.
        dpi (int, optional): The dots per inch of the saved quiver plot images.
            Defaults to 1000.
        mpp (float, optional): The micrometres per pixel value of the
            image sequence, if None will use pixel units. Defaults to None.
    """
    if mpp is None:
        mpp = 1
    for i, pred in enumerate(preds):
        x, y = pred[..., 0].flatten() * mpp, pred[..., 1].flatten() * mpp
        hist = polar(x, y)
        hist.plot.polar_map(cmap="rainbow")
        path = os.path.join(output_dir, 'Image_{}.png'.format(i))
        plt.savefig(path, dpi=dpi, bbox_inches=0)
        plt.close()



from setuptools import find_packages, setup

REQUIRED_PACKAGES = [
    'moviepy',
    'matplotlib',
    'natsort',
    'opencv-python',
    'pandas',
    'PIMS',
    'nd2reader',
    'PyYAML',
    'scikit-image',
    'tqdm',
    'ipympl',
    'einops',
    'pytest',
    'Gooey',
    'imutils',
    'astropy',
    'jupyterlab',
    'physt',
    'bumpver',
    'Pillow < 9.1.0 ; platform_system=="Darwin"'
]

CLASSIFIERS = [
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Operating System :: OS Independent',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: MacOS',
    'Operating System :: POSIX :: Linux',
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Science/Research',
    'Natural Language :: English',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Image Processing',
    'Topic :: Scientific/Engineering :: Visualization',
]

setup(
    name='rainbow-optical-flow',
    author='Alphons Gwatimba',
    author_email='0go0vdp95@mozmail.com',
    packages=find_packages(exclude=['tests*']) + ['misc'],
    version="2022.4.6",
    url='https://github.com/AlphonsG/Rainbow-Optical-Flow-For-ALI',
    license='MIT',
    classifiers=CLASSIFIERS,
    description=('Automated air liquid interface cell culture analysis using '
                 'deep optical flow.'),
    long_description=open('README.md', encoding='UTF-8').read(),
    long_description_content_type='text/markdown',
    install_requires=REQUIRED_PACKAGES,
    entry_points={
        'console_scripts': [
            'rainbow = rainbow.__main__:main'
        ]
    },
    python_requires='>=3.8',
    include_package_data=True,
    setup_requires=['wheel', 'setuptools']
)



from abc import ABCMeta, abstractmethod


class BaseModel:
    """Interface for optical flow models.

    Base class that optical flow model classes must inherit from.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        """Initializes class instance.

        Raises:
            NotImplementedError: Child class does not implement method.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_instance(model_cfg, reuse_model):
        """Returns an instance of the class.

        Args:
            model_cfg (dict): The configuration of the optical flow model.
            reuse_model (bool): If True, multiple calls of this method
            will return the same optical model instance.
            Defaults to True.

        Raises:
            NotImplementedError: Child class does not implement method.
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self):
        """Returns the optical flow of an image sequence as a list of
        numpy arrays of dimension [x, y, 2].

        Raises:
            NotImplementedError: Child class does not implement method.
        """
        raise NotImplementedError

    def get_img_pairs(self, imgs):
        """Returns image pairs.

        Duplicates every image of an image sequence, except the first and last,
        and groups pairs of images.

        Args:
            imgs (list): A list of images.

        Returns:
            list: A list of tuple images pairs.
        """
        img_pairs = []
        for i in range(0, len(imgs) - 1):
            img_pairs.append((imgs[i], imgs[i + 1]))

        return img_pairs



import torch
import torch.nn as nn
import torch.nn.functional as F

from update import GMAUpdateBlock
from extractor import BasicEncoder
from corr import CorrBlock
from utils.utils import bilinear_sampler, coords_grid, upflow8
from gma import Attention, Aggregate

try:
    autocast = torch.cuda.amp.autocast
except:
    # dummy autocast for PyTorch < 1.6
    class autocast:
        def __init__(self, enabled):
            pass

        def __enter__(self):
            pass

        def __exit__(self, *args):
            pass


class RAFTGMA(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128
        args.corr_levels = 4
        args.corr_radius = 4

        if 'dropout' not in self.args:
            self.args.dropout = 0

        # feature network, context network, and update block
        self.fnet = BasicEncoder(output_dim=256, norm_fn='instance', dropout=args.dropout)
        self.cnet = BasicEncoder(output_dim=hdim + cdim, norm_fn='batch', dropout=args.dropout)
        self.update_block = GMAUpdateBlock(self.args, hidden_dim=hdim)
        self.att = Attention(args=self.args, dim=cdim, heads=self.args.num_heads, max_pos_size=160, dim_head=cdim)

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img):
        """ Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8).to(img.device)
        coords1 = coords_grid(N, H // 8, W // 8).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """ Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination """
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, image1, image2, iters=12, flow_init=None, upsample=True, test_mode=False):
        """ Estimate optical flow between pair of frames """

        image1 = 2 * (image1 / 255.0) - 1.0
        image2 = 2 * (image2 / 255.0) - 1.0

        image1 = image1.contiguous()
        image2 = image2.contiguous()

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        with autocast(enabled=self.args.mixed_precision):
            fmap1, fmap2 = self.fnet([image1, image2])

        fmap1 = fmap1.float()
        fmap2 = fmap2.float()
        corr_fn = CorrBlock(fmap1, fmap2, radius=self.args.corr_radius)

        # run the context network
        with autocast(enabled=self.args.mixed_precision):
            cnet = self.cnet(image1)
            net, inp = torch.split(cnet, [hdim, cdim], dim=1)
            net = torch.tanh(net)
            inp = torch.relu(inp)
            # attention, att_c, att_p = self.att(inp)
            attention = self.att(inp)

        coords0, coords1 = self.initialize_flow(image1)

        if flow_init is not None:
            coords1 = coords1 + flow_init

        flow_predictions = []
        for itr in range(iters):
            coords1 = coords1.detach()
            corr = corr_fn(coords1)  # index correlation volume

            flow = coords1 - coords0
            with autocast(enabled=self.args.mixed_precision):
                net, up_mask, delta_flow = self.update_block(net, inp, corr, flow, attention)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_predictions.append(flow_up)

        if test_mode:
            return coords1 - coords0, flow_up

        return flow_predictions



import os
from multiprocessing import Process, Queue
from pathlib import Path

from rainbow import OPTICAL_FLOW_FILENAME
from rainbow.data_analysis import analyze_data
from rainbow.optical_flow.optical_flow import compute_optical_flow
from rainbow.util import load_nd2_imgs, load_std_imgs

from tqdm import tqdm

SENTINEL = 'STOP'


def process_files(root_dir, config, num_wrkrs, subdirs, overwrite_flow):
    """Processes files located within root_dir.

    Finds valid files (image sequences) within root_dir from which to compute
    the optical flow from and perform data analysis.

    Args:
        root_dir (string): The path to an existing directory containing image
            sequences as .nd2 files or 2 or more image files in common formats
            (.png, .tiff, etc).
        config (string): The path to a .yaml configuration file.
        num_wrkrs (int): The number of workers to use for parallel analysis.
            If set to None will use the same number of workers as CPU core
            count.
        subdirs (bool): If True, will look for image sequences in root_dir
            subfolders instead of top level directory.
        overwrite_flow (bool): If True, will compute optical flow
            even if a flow file already exists for an image sequence.
    """
    queue = Queue()
    if num_wrkrs != 1:
        wrkrs = initialize_workers(num_wrkrs, config, queue)

    try:
        dirs = [os.path.join(root_dir, curr_dir) for curr_dir in
                next(os.walk(root_dir))[1]] if subdirs else [root_dir]
    except StopIteration:
        print(f'Root directory ({root_dir}) is empty.')

    pbar = tqdm(total=1)
    for curr_dir in dirs:
        try:
            files = next(os.walk(curr_dir))[2]
        except StopIteration:
            files = []
        img_paths = [curr_dir] + [os.path.join(curr_dir, f) for f in files if
                                  Path(f).suffix == '.nd2']
        for img_path in img_paths:
            imgs = ([load_std_imgs(img_path, config['mpp'])] if
                    os.path.isdir(img_path) else load_nd2_imgs(img_path,
                    config['nd2'], config['mpp']))
            if len(imgs[0]) == 0:
                continue

            pbar.total += len(imgs)
            pbar.refresh()
            for img_seq in imgs:
                if len(img_seq) < 2:
                    pbar.update()
                    continue
                output_dir = get_output_dir(img_seq, config)
                if not skip_opt_flow(output_dir, overwrite_flow):
                    compute_optical_flow(img_seq, output_dir, config[
                        'opt_flow_model'], config[config['opt_flow_model']],
                        overwrite_flow=overwrite_flow)

                queue.put(output_dir)
                if num_wrkrs == 1:
                    queue.put(SENTINEL)
                    analyze_data(queue, config)
                    queue.get()

                pbar.update()

        if not subdirs:
            break

    if num_wrkrs != 1:
        queue.put(SENTINEL)
        for wrkr in wrkrs:
            wrkr.join()
        pbar.update()
    else:
        pbar.update()

    pbar.close()


def initialize_workers(num_wrkrs, config, queue):
    """Initializes workers.

    Creates and starts workers (subprocesses) that will perfrom data analysis.

    Args:
        num_wrkrs (int): The number of workers to create. If set to None will
        create the same number of workers as CPU core count.
        config (string): The path to a .yaml configuration file.
        queue (multiprocessing.Queue): The queue that will be used to
            communicate with the workers.

    Returns:
        list: A list of initalized workers as multiprocessing.Process objects.
    """
    if num_wrkrs is None:
        num_wrkrs = os.cpu_count() if os.cpu_count() is not None else 1

    wrkrs = []
    for i in range(0, num_wrkrs):
        wrkr = Process(target=analyze_data, args=(queue, config))
        wrkr.daemon = True
        wrkrs.append(wrkr)

    for wrkr in wrkrs:
        wrkr.start()

    return wrkrs


def get_output_dir(imgs, config):
    """Determines the corresponding output directory for an image sequence.

    The automatically named output directory will be used to store the
    data generated from an image sequence and is not created if
    nonexisting.

    Args:
        imgs (list): An image sequence as a list of PIMS Frame arrays.
        config (string): The path to a .yaml configuration file.

    Returns:
        string: The path to the output directory.
    """
    name, ext = os.path.splitext(imgs[0].metadata['img_name'])
    ser = ('_Series_{})'.format(imgs[0].metadata[config['nd2'][
           'naming_axs'][0]]) if ext == '.nd2' else '')
    output_dir = '{}_{}{}_etc'.format(ext.replace('.', ''), name, ser)
    output_dir = output_dir.replace(' ', '_')
    output_dir = os.path.join(imgs[0].metadata['img_ser_md']['dir'],
                              output_dir)

    return output_dir


def skip_opt_flow(output_dir, overwrite_flow):
    """Determines whether optical flow computation should be skipped.

    Skips computation iff an optical flow file already exists in output_dir
    and optical flow is not set to be overwritten.

    Args:
        output_dir (string): The path to an existing output directory.
        overwrite_flow (bool): If True, will ignore optical flow file if
            already existing in output_dir.

    Returns:
        bool: Whether optical flow computation should be skipped.
    """
    return True if not overwrite_flow and (Path(output_dir) / (
        OPTICAL_FLOW_FILENAME)).is_file() else False
